@@ -123,7 +123,7 @@ def ArchitectureParameter(**kwargs):
                         structure = Grammar.fromstring(structure)
                     else:
                         structure = ConstrainedGrammar.fromstring(structure)
-                        structure.set_constraints(**constraint_kwargs)
+                        structure.set_constraints(**constraint_kwargs)  # type: ignore[union-attr]
 
                 super().__init__(
                     grammar=structure,  # type: ignore[arg-type]
@@ -144,7 +144,12 @@ def ArchitectureParameter(**kwargs):
                 self.prune_graph()
 
                 if self._set_recursive_attribute:
-                    _build(self, self._set_recursive_attribute)
+                    m = _build(  # pylint: disable=assignment-from-no-return
+                        self, self._set_recursive_attribute
+                    )
+
+                if m is not None:
+                    return m
 
                 self.compile()
                 self.update_op_names()
